#!/bin/env perl -w
###################################################################
###                                                             ###
###      File: /projekte/CWB/tools/perl/CQP/lib/CQP.pm          ###
###   Version:                                                  ###
###    Author: Stefan Evert                                     ###
###   Purpose: Perl interface to CQP                            ###
###   Created: Tue Mar  2 17:58:17 1999                         ###
###  Modified: Fri Sep 26 14:42:39 2003 (evert)                 ###
###                                                             ###
###################################################################
#
# History:
#


package CQP;
$VERSION = '1.0';

# external packages used
use FileHandle;
use IPC::Open3;
use CWB;

# package global variables
$CQP = "$CWB::CQP -c";

sub new {
  my $class = shift;		# class name
  my $self = {};		# namespace for new CQP class object
  my $options = "@_";		# CQP command-line options (use at your own risk)

  # run CQP server in the background
  my $out = $self->{'out'} = new FileHandle;
  my $in = $self->{'in'} = new FileHandle;
  my $err = $self->{'err'} = new FileHandle;
  open3($out, $in, $err, "$CQP $options");
  $out->autoflush(1);

  my $version_string = $in->getline; # "cqp -c" should print version on startup
  chomp $version_string;
  die "ERROR CQP backend startup failed ('$CQP $options')\n"
    unless $version_string =~ /^CQP\s+(?:\w+\s+)*([0-9]+)\.([0-9]+)(?:\.b([0-9]+))?(?:\s+(.*))?$/;
  $self->{'major_version'} = $1;
  $self->{'minor_version'} = $2;
  $self->{'beta_version'} = $3;
  $self->{'compile_date'} = $4;
  if (not (			# we need cqp-2.2.b17 or newer (for query lock)
	   $1 >= 3
	   or ($1 == 2 and $2 == 2 and $3 >= 41)
	  )) {
    die "ERROR CQP version too old. ($version_string)\n";
  }

  # error handling
  $self->{'error_handler'} = ""; # set to subref for user-defined error handler
  $self->{'status'} = 'ok';	 # status of last executed command ('ok' or 'error')
  $self->{'error_message'} = []; # arrayref to array containing message produced by last error that occurred

  # debugging (prints more or less everything on stdout)
  $self->{'debug'} = 0;

  # stderr's select vector (in order to check error messages)
  my $bitvec = '';
  vec($bitvec, fileno($err), 1) = 1;
  $self->{'err_vec'} = $bitvec;

  # CQP object setup complete
  bless($self, $class);

  # ignore any output which may have been produced during startup
  $self->exec("set PrettyPrint off"); # pretty-printing should be turned off for non-interactive use

  return $self;
}

sub DESTROY {
  my $self = shift;

  my $out = $self->{'out'};
  if (defined $out) {
    $out->print("exit"); 	# exit CQP backend
    $out->close;	
  }
  my $in = $self->{'in'};
  if (defined $in) {
    $in->close;
  }
}

sub check_version {
  my $self = shift;
  my ($major, $minor, $beta) = @_;
  $minor = 0 unless defined $minor;
  $beta = 0 unless defined $beta;

  my $maj = $self->{'major_version'};
  my $min = $self->{'minor_version'};
  my $bet = $self->{'beta_version'};
  if ( $maj > $major
       or ($maj == $major and $min > $minor)
       or ($maj == $major and $min == $minor and $bet >= $beta)
     ) {
    return 1;
  }
  else {
    return 0;
  }
}

# execute CQP command
sub exec {
  my $self = shift;
  my $cmd = shift;
  my @result = ();
  my $line;
  my $debug = $self->{'debug'};

  $self->{'status'} = 'ok';
  chomp $cmd;			# make sure the user hasn't added a newline character
  $cmd =~ s/;\s*$//;		# ";" will be added when $cmd is sent to CQP
  print "CQP << $cmd;\n"
    if $debug;
  $self->{'out'}->print("$cmd; .EOL.;\n");
  while (defined ($line = $self->{'in'}->getline)) {
    chomp $line;
    $line =~ tr[\r][]d;		# ignore CR codes CQP may produce in PrettyPrint mode
    if ($line =~ /^-::-EOL-::-$/) {	# special line printed by ".EOL.;" marks end of command output
      print "CQP ", "-" x 60, "\n"
	if $debug;
      last;
    }
    print "CQP >> $line\n"
      if $debug;
    push @result, $line
      unless $line =~ /^$/;	# ignore end-of-command markers (i.e. empty lines)
  }
  $self->checkerr;		# check for error messages
  return @result;
}

# run query in safe mode (query lock)
sub query {
  my $self = shift;
  my $query = shift;
  my @result = ();
  my $key = int rand(1_000_000);
  my @errmsg = ();		# collect CQP error messages
  my $error = 0;		# check if any error occurs

  $self->exec("set QueryLock $key"); # enter query lock mode
  if ($self->status ne 'ok') {
    push @errmsg, $self->error_message;
    $error = 1;
  }
  @result = $self->exec($query);
  if ($self->status ne 'ok') {
    push @errmsg, $self->error_message;
    $error = 1;
  }
  $self->exec("unlock $key");	    # unlock with our random key
  if ($self->status ne 'ok') {
    push @errmsg, $self->error_message;
    $error = 1;
  }
  # now set error status & error message
  $self->{'status'}  = ($error) ? 'error' : 'ok';
  $self->{'error_message'} = \@errmsg;
  return @result;
}


# dump subcorpus
sub dump {
  my $self = shift;
  my $subcorpus = shift;
  my $from = shift;
  my $to = shift;
  my @matches = ();
  my $line;

  if ((not defined $subcorpus) or
      (defined $from and not defined $to)) {
    die "USAGE:  \$cqp->dump(\$named_query [, \$from, \$to]);\n";
  }
  $from = "" unless defined $from;
  $to = "" unless defined $to;

  my @result = $self->exec("dump $subcorpus $from $to");
  foreach $line (@result) {
    push @matches, [ split /\t/, $line ];
  }
  return @matches;
}

# check CQP's stderr stream for error messages
sub checkerr {
  my $self = shift;
  my $rvec = $self->{'err_vec'};

  my $ready = select($rvec, undef, undef, 0);
  if ($ready) {
    # we've got something on stderr -> an error must have occurred
    $self->{'status'} = 'error';
    my @msg = $self->readerr;
    $self->{'error_message'} = \@msg;
    $self->error("**** CQP ERROR ****", @msg);
  }
}

# read all available lines from CQP's stderr stream
sub readerr {
  my $self = shift;
  my $err  = $self->{'err'};
  my $rvec = $self->{'err_vec'};

  my $msg = "";
  my $block;
  while (select($rvec, undef, undef, 0)) {
    sysread $err, $block, 256;
    $msg .= $block;
  }
  return split /\n/, $msg;
}

# method to read the CQP object's status and error message variables
sub status {
  my $self = shift;

  return $self->{'status'};
}

# simplified interface for checking for CQP errors
sub ok {
  my $self = shift;
  return ($self->status eq 'ok');
}

sub error_message {
  my $self = shift;
  my $aref = $self->{'error_message'};

  return @{$aref};
}

# error messages (optionally through user-defined error handler)
sub error {
  my $self = shift;

  if (ref $self->{'error_handler'} eq 'CODE') {	# subref
    &{$self->{'error_handler'}}(@_);
  }
  else {
    foreach (@_) {
      print STDERR "$_\n";
    }
  }
}

# set user-defined error handler
sub set_error_handler {
  my $self = shift;
  my $handler = shift;

  die "USAGE:  \$cqp->set_error_handler(\\\&my_error_handler);\n"
    unless (not defined $handler) or (ref $handler eq 'CODE');
  $self->{'error_handler'} = $handler;
}
  
# switch debugging output on/off
sub debug {
  die "USAGE:  \$prev_status = \$cqp->debug( 1 | 0 ) ;\n"
    unless @_ == 2;
  my $self = shift;
  my $on = shift;
  my $prev = $self->{'debug'};
  $self->{'debug'} = $on;
  return $prev;
}


return 1;


__END__


=head1 NAME

  CQP - Create a CQP server process as a Perl object


=head1 SYNOPSIS

  use CQP;

  # start CQP server process in the background
  $my_cqp = new CQP;
  $my_cqp = new CQP("-r /home/britney/registry", "-I /global/init.cqp");

  # check for required minimum CQP version
  $ok = $my_cqp->check_version($major [, $minor [, $beta]]);

  # execute CQP command and check if an error has occurred
  @lines = $my_cqp->exec($my_cmd);
  if (not $my_cqp->ok) {
    @cqp_error_message = $my_cqp->error_message;
    my_error_handler();
  }

  # it's easier to use an automatic error handler
  $my_cqp->set_error_handler(\&my_error_handler);

  # execute in query lock mode (to improve security of CGI scripts)
  @lines = $my_cqp->query($my_query);

  # dump a named query and return regions as array of arrayrefs
  @matches = $my_cqp->dump("Last" [, $from, $to]);
  foreach $m (@matches) {
    ($start, $end, $target, $keyword) = @{$m};
  }

  # close down CQP server (exits gracefully)
  undef $my_cqp;


=head1 DESCRIPTION

When you create a I<CQP> object, CQP (the IMS Corpus Query Processor)
is run in the background. Every I<CQP> object has its own CQP server
process. A I<CQP> object can execute any CQP command, and returns the
output which CQP would normally print to C<STDOUT>.

Error detection is implemented by monitoring the server's C<STDERR>
stream. Error messages printed by CQP can be obtained through an
object method call.


=head1 METHODS

The following methods are available:

=over 4

=item $cqp = CQP->new;

Spawn new CQP background server. The object returned is used to
communicate with the server process.

=item $cqp = CQP->new('-r /home/britney/registry');

Optional arguments of the C<new> method are passed as command-line
options to CQP. Use at your own risk.

=item $ok = $cqp->check_version($major, $minor, $beta);

Check for minimum required CQP version, i.e. 
CQP-E<lt>$majorE<gt>.E<lt>$minorE<gt>.bE<lt>$betaE<gt>
$minor and $beta may be omitted, in which case they default to 0.

=item @results = $cqp->exec($cmd);

Execute a I<single> CQP command by sending the string $cmd to CQP's
input. This call will block until CQP has completed command execution.
Regular CQP output in response to the command will be returned as an
array. Each array element corresponds to one line of CQP output
(usually a I<kwic> line) with the trailing newline stripped. CQP error
messages are reported through the I<CQP> error handling mechanism (see
below). B<NB> If you execute multiple CQP commands in a single call,
you will mess up the Perl/CQP interface of this object.

=item @results = $cqp->query($cmd);

Execute $cmd in I<query lock> mode, where only queries can be
executed. Otherwise like $cqp-E<gt>exec(). This method is provided to
improve security in CGI and server applications.

=item @matches = $cqp->dump($named_query [, $from, $to]);

Dump query matches using the CQP 'dump' command. Each element of
@matches corresponds to one matching region. It is a reference to an
anonymous array containing the values

  (match_start, match_end, target, keyword).

Undefined keywords / targets are set to -1. If C<$from> and C<$to> are 
specified, only matches from that range are returned (don't forget that
matches are numbered starting with 0!).

=item $cqp->status;

Returns the status of the last I<exec>ed command, which is either 'ok'
or 'error'.

=item $cqp->ok;

Returns C<True> iff the status of the last I<exec>ed command is 'ok'.
This method is an abbreviation for the expression

  ($cqp->status eq 'ok')

=item @lines = $cqp->error_message;

Returns the last error reported by CQP. This is not reset
automatically, so you need to check I<$cqp-E<gt>status;> in order to
see if the error message was actually produced by the last command.
The return value is a list of error message lines with newlines stripped.

=item $cqp->error(@lines);

Used internally to report errors and CQP error messages. May be used
to add error conditions in the main program that are connected with a
given I<CQP> object to this object's error listing.

=item $cqp->set_error_handler(\&my_error_handler);

Set user-defined error handler callback function for this CQP process.
All internal error messages will be sent to the callback function,
which receives a list of error message lines as arguments. When
I<$cqp-E<gt>exec> or I<$cqp-E<gt>checkerr> detect an error, they will
effectively execute
  
  my_error_handler($cqp->error_message);

Calling I<$cqp-E<gt>set_error_handler(undef)> will restore the default
behaviour of echoing error messages on C<STDOUT>. If you create more
than one I<CQP> object, you can use closures over anonymous
subroutines to provide each server process with its own error handler.

=item undef $cqp;

Exits CQP gracefully by issuing an "exit;" command.

=back


=head1 VARIABLES

It should normally not be necessary to access the internal variables
of I<CQP> objects directly. Use methods instead. However, the
following variables may be of interest to the advanced user:

=over 4

=item $CQP::CQP

The system command used to create a CQP server process.

=item $cqp->{'major_version'}

=item $cqp->{'minor_version'}

=item $cqp->{'beta_version'}

=item $cqp->{'compile_date'}

Version information provided by the CQP server process.

=back


=head1 COPYRIGHT

Copyright 1999-2002 Stefan Evert (IMS Stuttgart).

CQP copyright 1993-2003 IMS Stuttgart.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
