# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

BEGIN { $| = 1; print "Running tests for CQP.pm:\n"; }

# step 1: load the CQP/Perl library
END {
  system "rm -f DICKENS:A*";			# clean up temporary corpus files
  if (not $loaded) {
    print "!! can't load CQP.pm (aborted)\n";
    exit 1;
  }
  if (not $deleted) {
    print "!! error deleting CQP object\n";
    exit 1;
  }
}
use CQP;
$loaded = 1;
print " - use CQP.pm ok\n";

# step 2: start CQP process with our own test init file
$CQP = new CQP ("-I ./init.cqp");
if (not defined $CQP) {
  print "!! can't start CQP process\n";
  exit 1;
}
print " - spawned CQP child process\n";

# step 3: we need an up-to-date version of CQP
$ok = $CQP->check_version(2, 2, 49);		# thats 2.2.b49 or newer
if (not $ok) {
  print "!! CQP version too old, consider upgrading\n";
  exit 1;
}
$major = $CQP->{'major_version'};
$minor = $CQP->{'minor_version'};
$beta =  $CQP->{'beta_version'};
$date =  $CQP->{'compile_date'};
print " - CQP version $major.$minor";
print "b$beta"
  if $beta > 0;
print " $date\n";

# step 4: list available corpora (run CQP manually with -c to check output of interactive commands!)
@corpora = $CQP->exec("show corpora");
map {s/^\s+//} @corpora;			# output of "show corpora;" is TAB-indented
shift @corpora;					# first line of output is "header"
if ((not $CQP->ok) or (@corpora == 0) or (not grep {$_ eq "DICKENS"} @corpora)) {
  print "!! error in 'show corpora;' command\n";
  exit 1;
}
$N = @corpora;
print " - $N corpora found, including DICKENS\n";

# step 5: activate corpus
$CQP->exec("DICKENS");
if (not $CQP->ok) {
  print "!! can't activate corpus DICKENS\n";
  exit 1;
}
print " - corpus DICKENS activated\n";

# step 6: we need a little trick to get at the corpus size
@matches = $CQP->dump("DICKENS");		# dont ask ...
if (@matches != 1) {
  print "!! error in 'dump DICKENS;' command\n";
  exit 1;
}
$size = $matches[0]->[1] + 1;			# I told you not to ask ...
if (not $size > 1_000_000) {
  print "!! error reading corpus size\n";
  exit 1;
}
print " - corpus size = $size tokens\n";

# step 7: try an error condition, but dont let the default error handler mess up the screen
$CQP->set_error_handler(sub { return 0 });
$CQP->exec("[lemma == ']");
if (CQP->ok) {
  print "!! query syntax error not detected\n";
  exit 1;
}
@msg = $CQP->error_message;
if (not ((@msg >= 2) and grep {/<--/} @msg)) {	# the <-- is used by CQP to indicate the location of a syntax error
  print "!! incomplete error message\n";
  exit 1;
}
$N = @msg;
print " - syntax error detected ($N-line error message)\n";

# step 8: try a different error condition and check whether our custom error handler is really called
$handler_ok = 0;
$CQP->set_error_handler(sub { if (@_ >= 2) { $handler_ok = @_ } } );
$CQP->exec("'('");
if (not $handler_ok) {
  print "!! custom error handler was not invoked\n";
  exit 1;
}
print " - syntax error handled ($handler_ok-line error message)\n";

# step 8b: try a query lock evaluation with an interactive command
$handler_ok = 0;
@nothing = $CQP->query("show cd");
if (not ($handler_ok and @nothing == 0)) {
  print "!! query lock violation was not detected\n";
  exit 1;
}
print " - query lock violation caught\n";

# set stricter handler (any further CQP errors are _real_ errors ...)
$CQP->set_error_handler(sub {
			  print "\n!! CQP execution error\n";
			  exit 1;
			 } );

# step 9: check multi-line commands
@nothing = $CQP->exec(";;;;;");			# five blank lines as end-of-command markers are swallowed by CQP.pm
if (@nothing > 0) {
  print "!! unexpected output from empty comands\n";
  exit 1;
}
print " - command ';;;;;' ok, no output\n";

# step 10: repeatedly run a macro query defined in init.cqp and save the results as named queries
print " - executing macro ";
for ($i = 0; $i < 8; $i++) {
  print ".";
  $CQP->exec("A$i = /example[$i]");
}
print " ok\n";

# step 11: save results (named queries) to disk

print " - saving named queries ";
for ($i = 0; $i < 8; $i++) {
  print ".";
  $CQP->exec("save A$i");
}
print " ok\n";
@saved = glob "DICKENS:A*";
if (@saved != 8) {
  print "!! can't find saved query results on disk\n";
  exit 1;
}

# step 12: exit and re-start CQP
undef $CQP;
$deleted = 1;
print " - CQP process terminated\n";
$CQP = new CQP ("-I ./init.cqp");
if (not defined $CQP) {
  print "!! can't re-start CQP process\n";
  exit 1;
}
print " - new CQP child process spawned\n";

# step 13: named queries from step 10 should now be visible within CQP
@named = $CQP->exec("show named");
shift @named;					# skip "header" line
if ((grep {/-d-\s+DICKENS:A/} @named) != 8) {
  print "!! wrong output from 'show named;' command\n";
  exit 1;
}
print " - saved query results found\n";

# step 14: re-load query results and get number of matches
@size = ();
$total_size = 0;				# should add up to number of sentences in DICKENS
print " - loading query results ";
for ($i = 0; $i < 8; $i++) {
  print ".";
  ($s) = $CQP->exec("size A$i");		# "size" command returns just one number 
  if (not (defined $s and $s >= 0)) {
    print "\n!! load error\n";
    exit 1;
  }
  push @size, $s;
  $total_size += $s;
}
print " ok\n";
if (not ($size[5] > 0 and $size[7] == 0)) {
  print "!! wrong number of matches\n";
  exit 1;
}
print " - number of matches: ", join(", ", @size), "\n";

# step 15: check $total_size against simple query counting all sentences
($size) = $CQP->exec("DICKENS; A = <s>[]; size A;"); # can execute multiple command with single call
if ($total_size != $size) {
  print "!! inconsistency in number of matches ($size vs. $total_size)\n";
  exit 1;
}
print " - total: $size sentences\n";

# step 16: cat the query results from A4 and A5 and check the number of lines
$CQP->exec("show -cpos");			# don't print cpos (check absence below)
@lines4 = $CQP->exec("cat A4");
@lines5 = $CQP->exec("cat DICKENS:A5");		# use both short and full form of named query
if (not (@lines4 == $size[4] and @lines5 == $size[5])) {
  print "!! 'cat;' command failed\n";
  exit 1;
}
if (grep {/^\s*[0-9]+:/ or not /<.+>/} (@lines4, @lines5)) {
  print "!! error in 'cat;' output format\n";
  exit 1;
}
print " - 'cat;' command ok, first match with 5 NPs: \n";
print "   ``$lines5[0]''\n";

# step 17: set target and keyword (trivially) ...
print " - executing query ... ";
$CQP->exec("A = [pos = 'DT'] [pos = 'JJ.*']* @[pos = 'NNS?'] within s");
print "set keyword ... ";
$CQP->exec("set A keyword nearest [pos = 'DT'] within left 1 s from matchend");
print "ok\n";
($size, $size_target, $size_keyword) =
  $CQP->exec("size A; size A target; size A keyword;"); # isnt that just neat?
if (not ($size == $size_target and $size == $size_keyword)) {
  print "!! wrong number of targets/keywords\n";
  exit 1;
}
print " - $size matches with target and keyword anchors\n";

# step 18: ... and get them back from the dump command
print " - checking first 50,000 matches ..";
@matches = $CQP->dump("A", 0, 49_999);
if (@matches != 50_000) {
  print "\n!! wrong number of matches in dump\n";
  exit 1;
}
print "..";
foreach $m (@matches) {
  ($match, $matchend, $target, $keyword) = @$m;
  if ($match != $keyword or $matchend != $target) {
    print "\n!! anchors incorrectly set (", join(", ", @$m), ")\n";
    exit 1;
  }
}
print " ok\n";

# that's it -- we've passed the test
print "Congratulations. All tests passed.\n";
exit 0;



